function register_FISSEQ_images(input_dir,output_dir,nbins,overlap,subpixel_precision)


% Evan Daugharthy
% Genetics Department, Harvard University
% Wyss Institute for Biologically Inspired Engineering
% 2014

%
% Validate Input
%

% Initialize timer
tic;
start_time = clock;

% Check that the input directory exists
if exist(input_dir) ~= 7
    error('Specified Input Directory does not exist')
end

% Create the output directory if it does not exist
if exist(output_dir) == 0
    mkdir(output_dir)
end

% Find .ics files in the directory
files = dir(input_dir);
ics_files = regexp({files.name},'/*.ics','match');
w_ics_files = find(~cellfun(@isempty,ics_files));

% Default subpixel_precision to 1
if exist('subpixel_precision')~=1, subpixel_precision=1; end

% Default nbins to 10
if exist('nbins')~=1, nbins=10; end

% Default overlap to 0.1
if exist('overlap')~=1, overlap=0.1; end


% Report time
disp('Initialize:');
toc;

%
% Initialize data array
%

% Initialize timer
tic;

% Determine the number of bases
for i = 1:length(w_ics_files)
    datsize = strsplit(files(w_ics_files(i)).name,'_');
    basenum(i) = str2num(datsize{1});
end
basenum = max(basenum);

% Preallocate storage to keep track
dathold{basenum+1} = 'storage';
where_dathold{basenum+1} = 'where_storage';
hold = struct([]);

% Process the first file in the list and pre-allocate storage
im = bfopen(strcat(input_dir, '/',files(w_ics_files(1)).name));
% Find the number of stacks
stacks = size(im{1,1},1);
% Save and parse the metadata to get image dimensions and file name
metadata = im{1,2};
layout_order = metadata.get('Global layout order');
filename = metadata.get('Global filename');
layout_size = metadata.get('Global layout sizes');
mapObj = containers.Map(strsplit(layout_order,' '),strsplit(layout_size,' '));
x = str2num(mapObj('x'));
y = str2num(mapObj('y'));
z = str2num(mapObj('z'));
filename_split = strsplit(filename,{'\','/'});
basestring = filename_split{size(filename_split,2)};
basestring = strsplit(basestring,'_');
base = str2num(basestring{1});
% pre-allocate storage
data = zeros([y,x,5,basenum]);
im_data = zeros(y,x,z,5);
for j = 1:stacks
    plane_md = im{1,1}{j,2};
    % Find Z
    Zexpr = ['Z=' '(.*?)' '/'];
    Ztok = regexp(plane_md,Zexpr,'tokens');
    Z = str2num(Ztok{1}{1});
    % Find Ch
    Cexpr = ['C=' '(.*?)' '/'];
    Ctok = regexp(plane_md,Cexpr,'tokens');
    C = str2num(Ctok{1}{1});
    % Place the Z-slice and store localization data
    im_data(:,:,Z,C) = im{1,1}{j,1};
end
im_data(:,:,:,5) = max(im_data(:,:,:,1:4),[],4);
dathold{1} = squeeze(max(im_data,[],3));
where_dathold{1} = base;

% Import image stack for each timepoint
parfor i = 2:length(w_ics_files)
    im = bfopen(strcat(input_dir, '/',files(w_ics_files(i)).name));
    % Find the number of stacks
    stacks = size(im{1,1},1);
    % Save and parse the metadata to get image dimensions and file name
    metadata = im{1,2};
    layout_order = metadata.get('Global layout order');
    filename = metadata.get('Global filename');
    layout_size = metadata.get('Global layout sizes');
    mapObj = containers.Map(strsplit(layout_order,' '),strsplit(layout_size,' '));
    x = str2num(mapObj('x'));
    y = str2num(mapObj('y'));
    z = str2num(mapObj('z'));
    filename_split = strsplit(filename,{'\','/'});
    basestring = filename_split{size(filename_split,2)};
    basestring = strsplit(basestring,'_');
    base = str2num(basestring{1});
    % pre-allocate storage
    im_data = zeros(y,x,z,5);
    for j = 1:stacks
        plane_md = im{1,1}{j,2};
        % Find Z
        Zexpr = ['Z=' '(.*?)' '/'];
        Ztok = regexp(plane_md,Zexpr,'tokens');
        Z = str2num(Ztok{1}{1});
        % Find Ch
        Cexpr = ['C=' '(.*?)' '/'];
        Ctok = regexp(plane_md,Cexpr,'tokens');
        C = str2num(Ctok{1}{1});
        % Place the Z-slice and store localization data
        im_data(:,:,Z,C) = im{1,1}{j,1};
    end
    im_data(:,:,:,5) = max(im_data(:,:,:,1:4),[],4);
    dathold{i} = squeeze(max(im_data,[],3));
    where_dathold{i} = base;
end

% Assemble images from container
for i = 1:length(w_ics_files)
    data(:,:,:,where_dathold{i}) = dathold{i};
end

% Clear memory hog storage containers
clear dathold;
clear im_data;

% Report time
disp('Import & Store All Images:');
disp(sprintf('Image Data (%d,%d,5,%d)',x,y,basenum));
toc;

%
% Image Registration
%

% Initialize timer
tic;

% Make data 8-bit
data = im2uint8(mat2gray(data));

% Process Alignment
[Routput, Rchadj, Rtadj, Regstack] = FISSEQ_registration(data,subpixel_precision,nbins,overlap);

% Write Tiff Images to file
for t = 1:size(Regstack,4)
    for ch = 1:size(Regstack,3)
        fn = sprintf('%sf3_T_%02d_C_%02d.tif',output_dir,t-1,ch-1);
        imwrite(im2uint8(mat2gray(Regstack(:,:,ch,t))),fn,'tiff');
    end
end

% Save alignment parameters
fn_Rout = sprintf('%sRoutput.mat',output_dir);
fn_Rchadj = sprintf('%sRchadj.mat',output_dir);
fn_Rtadj = sprintf('%sRtadj.mat',output_dir);
save(fn_Rout,'Routput');
save(fn_Rchadj,'Rchadj');
save(fn_Rtadj,'Rtadj');

% Report time
disp('Write Registered Images to File:')
toc;

% Report time
end_time = clock;
disp('Register Images:');
disp(sprintf('Elapsed time is %f seconds.',etime(end_time,start_time)))

return

function [Routput, Rchadj, Rtadj, Regstack] = FISSEQ_registration(im_data,usfac,nblocks,ol)

% function [Routput, Rchadj, Rtadj, Regstack] = FISSEQ_dftregistration_chadjust_blockreg2(im_data,usfac,nblocks,ol);
% Efficient blockwise image registration of FISSEQ Images
%
%
% Evan Daugharthy
% Genetics Department, Harvard University
% Wyss Institute for Biologically Inspired Engineering
% 2013
%
%

% Inputs
% im_data   4D Image Stack (Y, X, Ch, T)
% usfac     Upsampling factor (integer). Images will be registered to
%           within 1/usfac of a pixel. For example usfac = 20 means the
%           images will be registered within 1/20 of a pixel. (default = 1)
% nblocks   Number of blocks to divide the image into along each axis.
%           For example, nblocks=2 will give 4 total blocks (2x2)
% ol        Overlap between neighboring blocks. For example, 0.1 gives
%           a 10% overlap.
%

% Outputs
% Routput   Blockwise registration offsets over time. A matrix of size
%           (nblocks, nblocks, # T - 1, 4). The third index gives the 
%           offsets for each timepoint compared to the first time point.
%           The fourth index gives the dftregistration output for each
%           blockwise registration (see below).
% Rchadj    Blockwise registration offsets between channels (chromatic 
%           shifts). A matrix of size (nblocks, nblocks, # Ch - 1, 4).
%           The third index gives the offsets for each of channels 2, 3, 
%           and 4 compared to channel 1.
%           The fourth index gives the dftregistration output for each
%           blockwise registration (see below).
% Rtadj     Registration offsets over time for the whole image (not 
%           blockwise) as a rough initial registration attempt. A matrix of
%           size (# T - 1, 4). The first index gives the offsets for each 
%           time point compared to the first timepoint. The second index 
%           gives the dftregistration output for each blockwise 
%           registration (see below).
% Regstack  Registered 4D Image Stack (Y, X, Ch, T)
%


%
% This function is taken from dftregistration, which is cited as follows:
%
% function [output Greg] = dftregistration(buf1ft,buf2ft,usfac);
% Efficient subpixel image registration by crosscorrelation. This code
% gives the same precision as the FFT upsampled cross correlation in a
% small fraction of the computation time and with reduced memory 
% requirements. It obtains an initial estimate of the crosscorrelation peak
% by an FFT and then refines the shift estimation by upsampling the DFT
% only in a small neighborhood of that estimate by means of a 
% matrix-multiply DFT. With this procedure all the image points are used to
% compute the upsampled crosscorrelation.
% Manuel Guizar - Dec 13, 2007
% Portions of this code were taken from code written by Ann M. Kowalczyk 
% and James R. Fienup. 
% J.R. Fienup and A.M. Kowalczyk, "Phase retrieval for a complex-valued 
% object by using a low-resolution image," J. Opt. Soc. Am. A 7, 450-458 
% (1990).

% Citation for this algorithm:
% Manuel Guizar-Sicairos, Samuel T. Thurman, and James R. Fienup, 
% "Efficient subpixel image registration algorithms," Opt. Lett. 33, 
% 156-158 (2008).

% Inputs
% buf1ft    Fourier transform of reference image, 
%           DC in (1,1)   [DO NOT FFTSHIFT]
% buf2ft    Fourier transform of image to register, 
%           DC in (1,1) [DO NOT FFTSHIFT]
% usfac     Upsampling factor (integer). Images will be registered to 
%           within 1/usfac of a pixel. For example usfac = 20 means the
%           images will be registered within 1/20 of a pixel. (default = 1)

% Outputs
% output =  [error,diffphase,net_row_shift,net_col_shift]
% error     Translation invariant normalized RMS error between f and g
% diffphase     Global phase difference between the two images (should be
%               zero if images are non-negative).
% netre_row_shift net_col_shift   Pixel shifts between images
% Greg      (Optional) Fourier transform of registered version of buf2ft,
%           the global phase difference is compensated for.

%
% Begin Code
%

% Compute matrix size
[m,n,nch,b] = size(im_data);

% Compute bin sizes
stepsize_m = floor(m./nblocks);
mod_m = mod(m,nblocks);
stepsize_n = floor(n./nblocks);
mod_n = mod(n,nblocks);

% Compute overlap number of pixels
if nblocks > 1
    overlap_m = floor(stepsize_m.*ol);
    overlap_n = floor(stepsize_n.*ol);
elseif nblocks == 1
    overlap_m = 0;
    overlap_n = 0;
end

% Pre-allocate storage
im_data_tproj = zeros(m,n,4);
tempstore = zeros(m,n,nch,b,2);
Regstack = zeros(m,n,nch,b);
Rchadj = zeros(nblocks,nblocks,3,4);
Routput = zeros(nblocks,nblocks,b-1,4);
Rtadj = zeros(b-1,4);
bufft = zeros(m,n,nch,b);

% Calculate overall fourier transforms
for t = 1:b
    for ch = 1:nch
        bufft(:,:,ch,t) = fft2(im_data(:,:,ch,t));
    end
end

% Compute initial overall registration
buf1ft = bufft(:,:,5,1);
for t = 2:b
    buf2ft = bufft(:,:,5,t);
    output = fdftregistration(buf1ft,buf2ft,usfac);
    Regstack(:,:,:,t) = dftapply(bufft(:,:,:,t),output);
    % Store output
    Rtadj(t-1,:) = output;
end

% Pass through first timepoint
Regstack(:,:,:,1) = im_data(:,:,:,1);

% Compute max projection over time
im_data_tproj = max(Regstack(:,:,1:4,:),[],4);

% Locally adjust for chromatic shifts
for ch = 2:4
    for i = 1:nblocks
        for j = 1:nblocks
            % Store m block range
            if i == 1
                bin_m = 1:(stepsize_m + overlap_m);
            elseif i == nblocks
                bin_m = (stepsize_m.*(i-1) - overlap_m):(stepsize_m.*i + mod_m);
            else
                bin_m = (stepsize_m.*(i-1) - overlap_m):(stepsize_m.*i + overlap_m);
            end
            % Store n block range
            if j == 1
                bin_n = 1:(stepsize_n + overlap_n);
            elseif j == nblocks
                bin_n = (stepsize_n.*(j-1) - overlap_n):(stepsize_n.*j + mod_n);
            else
                bin_n = (stepsize_n.*(j-1) - overlap_n):(stepsize_n.*j + overlap_n);
            end
            % Compute fourier transforms
            buf1ft = fft2(im_data_tproj(bin_m,bin_n,1));
            buf2ft = fft2(im_data_tproj(bin_m,bin_n,ch));
            bufft = zeros(size(bin_m,2),size(bin_n,2),b);
            for t = 1:b
                bufft(:,:,t) = fft2(Regstack(bin_m,bin_n,ch,t));
            end
            % Compute chromatic offset for block
            output = fdftregistration(buf1ft,buf2ft,usfac);
            % Store output
            Rchadj(i,j,ch-1,:) = output;
            % Return registered subunit
            reg_im = dftapply(bufft,output);
            % Store
            tempstore(bin_m,bin_n,ch,:,2) = reg_im;
            tempstore(bin_m,bin_n,ch,:,1) = max(tempstore(bin_m,bin_n,ch,:,:),[],5);
        end
    end
end

% Pass through first channel
tempstore(:,:,1,:,1) = Regstack(:,:,1,:);

% Re-compute max projection
tempstore(:,:,5,:,:) = max(tempstore,[],3);

% Store channel registered data
Regstack = tempstore(:,:,:,:,1);

% Clear tempstore
tempstore = tempstore.*0;

% Calculate shifts over timepoints
for t = 2:b
    % Loop through each block generating the correct subset and registration
    for i = 1:nblocks
        for j = 1:nblocks
            % Store m block range
            if i == 1
                bin_m = 1:(stepsize_m + overlap_m);
            elseif i == nblocks
                bin_m = (stepsize_m.*(i-1) - overlap_m):(stepsize_m.*i + mod_m);
            else
                bin_m = (stepsize_m.*(i-1) - overlap_m):(stepsize_m.*i + overlap_m);
            end
            % Store n block range
            if j == 1
                bin_n = 1:(stepsize_n + overlap_n);
            elseif j == nblocks
                bin_n = (stepsize_n.*(j-1) - overlap_n):(stepsize_n.*j + mod_n);
            else
                bin_n = (stepsize_n.*(j-1) - overlap_n):(stepsize_n.*j + overlap_n);
            end
            % Compute fourier transforms
            buf1ft = fft2(Regstack(bin_m,bin_n,5,1));
            buf2ft = zeros(size(bin_m,2),size(bin_n,2),nch);
            for ch = 1:nch
                buf2ft(:,:,ch) = fft2(Regstack(bin_m,bin_n,ch,t));
            end
            % Compute offset for block
            output = fdftregistration(buf1ft,buf2ft(:,:,5),usfac);
            % Store output
            Routput(i,j,t-1,:) = output;
            % Return registered subunit
            reg_im = dftapply(buf2ft,output);
            % Store
            tempstore(bin_m,bin_n,:,t,2) = reg_im;
            tempstore(bin_m,bin_n,:,t,1) = max(tempstore(bin_m,bin_n,:,t,:),[],5);
        end
    end
end

% Pass through the first image
tempstore(:,:,:,1,1) = Regstack(:,:,:,1);

% Store channel registered data
Regstack = tempstore(:,:,:,:,1);

function output = fdftregistration(buf1ft,buf2ft,usfac)
% usfac
% size(buf1ft)
% Default usfac to 1
if exist('usfac')~=1, usfac=1; end

% Compute error for no pixel shift
if usfac == 0,
    CCmax = sum(sum(buf1ft.*conj(buf2ft)));
    rfzero = sum(abs(buf1ft(:)).^2);
    rgzero = sum(abs(buf2ft(:)).^2);
    error = 1.0 - CCmax.*conj(CCmax)/(rgzero*rfzero);
    error = sqrt(abs(error));
    diffphase=atan2(imag(CCmax),real(CCmax));
    output=[error,diffphase];

    % Whole-pixel shift - Compute crosscorrelation by an IFFT and locate the
    % peak
elseif usfac == 1,
    [m,n]=size(buf1ft);
    CC = ifft2(buf1ft.*conj(buf2ft));
    [max1,loc1] = max(CC);
    [~,loc2] = max(max1);
    rloc=loc1(loc2);
    cloc=loc2;
    CCmax=CC(rloc,cloc);
    rfzero = sum(abs(buf1ft(:)).^2)/(m*n);
    rgzero = sum(abs(buf2ft(:)).^2)/(m*n);
    error = 1.0 - CCmax.*conj(CCmax)/(rgzero(1,1)*rfzero(1,1));
    error = sqrt(abs(error));
    diffphase=atan2(imag(CCmax),real(CCmax));
    md2 = fix(m/2);
    nd2 = fix(n/2);
    if rloc > md2
        row_shift = rloc - m - 1;
    else
        row_shift = rloc - 1;
    end

    if cloc > nd2
        col_shift = cloc - n - 1;
    else
        col_shift = cloc - 1;
    end
    output=[error,diffphase,row_shift,col_shift];

    % Partial-pixel shift
else

    % First upsample by a factor of 2 to obtain initial estimate
    % Embed Fourier data in a 2x larger array
    %size(buf1ft)
    [m,n]=size(buf1ft);
    mlarge=m*2;
    nlarge=n*2;
    %m
    %n
    %mlarge
    %nlarge
    %size(buf1ft);
    CC=zeros(mlarge,nlarge);
    %size(CC)
    %size(CC(m+1-fix(m/2):m+1+fix((m-1)/2),n+1-fix(n/2):n+1+fix((n-1)/2)))
    k = fftshift(buf1ft).*conj(fftshift(buf2ft));
    %size(k)
    CC(m+1-fix(m/2):m+1+fix((m-1)/2),n+1-fix(n/2):n+1+fix((n-1)/2)) = k;

    % Compute crosscorrelation and locate the peak
    CC = ifft2(ifftshift(CC)); % Calculate cross-correlation
    [max1,loc1] = max(CC);
    [~,loc2] = max(max1);
    rloc=loc1(loc2);cloc=loc2;
    CCmax=CC(rloc,cloc);

    % Obtain shift in original pixel grid from the position of the
    % crosscorrelation peak
    [m,n] = size(CC); md2 = fix(m/2); nd2 = fix(n/2);
    if rloc > md2
        row_shift = rloc - m - 1;
    else
        row_shift = rloc - 1;
    end
    if cloc > nd2
        col_shift = cloc - n - 1;
    else
        col_shift = cloc - 1;
    end
    row_shift=row_shift/2;
    col_shift=col_shift/2;

    % If upsampling > 2, then refine estimate with matrix multiply DFT
    if usfac > 2,
        %%% DFT computation %%%
        % Initial shift estimate in upsampled grid
        row_shift = round(row_shift*usfac)/usfac;
        col_shift = round(col_shift*usfac)/usfac;
        dftshift = fix(ceil(usfac*1.5)/2); %% Center of output array at dftshift+1
        % Matrix multiply DFT around the current shift estimate
        CC = conj(dftups(buf2ft.*conj(buf1ft),ceil(usfac*1.5),ceil(usfac*1.5),usfac,...
            dftshift-row_shift*usfac,dftshift-col_shift*usfac))/(md2*nd2*usfac^2);
        % Locate maximum and map back to original pixel grid
        [max1,loc1] = max(CC);
        [~,loc2] = max(max1);
        rloc = loc1(loc2); cloc = loc2;
        CCmax = CC(rloc,cloc);
        rg00 = dftups(buf1ft.*conj(buf1ft),1,1,usfac)/(md2*nd2*usfac^2);
        rf00 = dftups(buf2ft.*conj(buf2ft),1,1,usfac)/(md2*nd2*usfac^2);
        rloc = rloc - dftshift - 1;
        cloc = cloc - dftshift - 1;
        row_shift = row_shift + rloc/usfac;
        col_shift = col_shift + cloc/usfac;

        % If upsampling = 2, no additional pixel shift refinement
    else
        rg00 = sum(sum( buf1ft.*conj(buf1ft) ))/m/n;
        rf00 = sum(sum( buf2ft.*conj(buf2ft) ))/m/n;
    end
    error = 1.0 - CCmax.*conj(CCmax)/(rg00*rf00);
    error = sqrt(abs(error));
    diffphase=atan2(imag(CCmax),real(CCmax));
    % If its only one row or column the shift along that dimension has no
    % effect. We set to zero.
    if md2 == 1,
        row_shift = 0;
    end
    if nd2 == 1,
        col_shift = 0;
    end
    output=[error,diffphase,row_shift,col_shift];
end

return

function reg_im = dftapply(bufft,output)
% Set up variables
diffphase = output(2);
row_shift = output(3);
col_shift = output(4);
[nr,nc,n_apply]=size(bufft);

% Preallocate array
reg_im = zeros(size(bufft));

% Compute registered version of the image stack
for it = 1:n_apply
    Nr = ifftshift([-fix(nr/2):ceil(nr/2)-1]);
    Nc = ifftshift([-fix(nc/2):ceil(nc/2)-1]);
    [Nc,Nr] = meshgrid(Nc,Nr);
    Greg = bufft(:,:,it).*exp(1i*2*pi*(-row_shift*Nr/nr-col_shift*Nc/nc));
    Greg = Greg*exp(1i*diffphase);
    reg_im(:,:,it) = abs(ifft2(Greg));
end

% Set pixels wrapped during registration to 0
if output(3) > 0
    reg_im(1:ceil(output(3)),:,:) = 0;
else
    reg_im(nr+floor(output(3)):nr,:,:) = 0;
end
if output(4) > 0
    reg_im(:,1:ceil(output(4)),:) = 0;
else
    reg_im(:,nc+floor(output(4)):nc,:) = 0;
end

return

function out=dftups(in,nor,noc,usfac,roff,coff)
% function out=dftups(in,nor,noc,usfac,roff,coff);
% Upsampled DFT by matrix multiplies, can compute an upsampled DFT in just
% a small region.
% usfac         Upsampling factor (default usfac = 1)
% [nor,noc]     Number of pixels in the output upsampled DFT, in
%               units of upsampled pixels (default = size(in))
% roff, coff    Row and column offsets, allow to shift the output array to
%               a region of interest on the DFT (default = 0)
% Recieves DC in upper left corner, image center must be in (1,1) 
% Manuel Guizar - Dec 13, 2007
% Modified from dftus, by J.R. Fienup 7/31/06

% This code is intended to provide the same result as if the following
% operations were performed
%   - Embed the array "in" in an array that is usfac times larger in each
%     dimension. ifftshift to bring the center of the image to (1,1).
%   - Take the FFT of the larger array
%   - Extract an [nor, noc] region of the result. Starting with the 
%     [roff+1 coff+1] element.

% It achieves this result by computing the DFT in the output array without
% the need to zeropad. Much faster and memory efficient than the
% zero-padded FFT approach if [nor noc] are much smaller than [nr*usfac nc*usfac]

[nr,nc]=size(in);
% Set defaults
if exist('roff')~=1, roff=0; end
if exist('coff')~=1, coff=0; end
if exist('usfac')~=1, usfac=1; end
if exist('noc')~=1, noc=nc; end
if exist('nor')~=1, nor=nr; end
% Compute kernels and obtain DFT by matrix products
kernc=exp((-1i*2*pi/(nc*usfac))*( ifftshift([0:nc-1]).' - floor(nc/2) )*( [0:noc-1] - coff ));
kernr=exp((-1i*2*pi/(nr*usfac))*( [0:nor-1].' - roff )*( ifftshift([0:nr-1]) - floor(nr/2)  ));
out=kernr*in*kernc;
return
